""" A kernel transformation that precomputes quantities until a vector register
is filled and then does vector computations """

from dune.perftool.loopy.vcl import get_vcl_type_size
from dune.perftool.loopy.transformations.vectorview import (add_temporary_with_vector_view,
                                                            add_vector_view,
                                                            get_vector_view_name,
                                                            )
from dune.perftool.tools import get_pymbolic_basename

from loopy.kernel.creation import parse_domains
from loopy.symbolic import pw_aff_to_expr

from pymbolic.mapper.dependency import DependencyMapper
from pymbolic.mapper.substitutor import substitute

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def collect_vector_data_rotate(knl, insns, inames):
    #
    # Process/Assert/Standardize the input
    #

    # inames input -> tuple
    if isinstance(inames, str):
        inames = inames.split(",")
    inames = tuple(i.strip() for i in inames)

    # insns -> list of Instruction instances
    if isinstance(insns, lp.match.MatchExpressionBase):
        insns = lp.find_instructions(knl, insns)
    else:
        if isinstance(insns, str):
            insns = [i.strip() for i in insns.split(",")]
        insns = [knl.id_to_insn[i] for i in insns]

    # Analyse the inames of the given instructions and identify inames
    # that they all have in common. Those inames will also be iname dependencies
    # of inserted instructions.
    common_inames = frozenset([]).union(*(insn.within_inames for insn in insns)) - frozenset(inames)

    # Determine the vector lane width
    # TODO infer the numpy type here
    vec_size = get_vcl_type_size(np.float64)

    # Add an iname to the kernel which will be used for vectorization
    new_iname = "quad_vec_{}".format("_".join(inames))
    domain = "{{ [{0}] : 0<={0}<{1} }}".format(new_iname, str(vec_size))
    domain = parse_domains(domain, {})
    knl = knl.copy(domains=knl.domains + domain)
    knl = lp.tag_inames(knl, [(new_iname, "vec")])

    new_insns = []
    all_writers = []

    #
    # Inspect the given instructions for dependent quantities
    #

    quantities = {}
    for insn in insns:
        for expr in DependencyMapper()(insn.expression):
            basename = get_pymbolic_basename(expr)
            quantities.setdefault(basename, frozenset())
            quantities[basename] = quantities[basename].union(frozenset([expr]))

    # Add vector size buffers for all these quantities
    replacemap_vec = {}
    for quantity in quantities:
        # Check whether there is an instruction that writes this quantity within
        # the given inames. If so, we need a buffer array.
        iname_match = lp.match.And(tuple(lp.match.Iname(i) for i in inames))
        write_match = lp.match.Writes(quantity)
        match = lp.match.And((iname_match, write_match))
        write_insns = lp.find_instructions(knl, match)
        all_writers.extend([i.id for i in write_insns])

        if write_insns:
            # Determine the shape of the quantity
            shape = knl.temporary_variables[quantity].shape

            arrname = quantity + '_buffered_arr'
            knl = add_temporary_with_vector_view(knl,
                                                 arrname,
                                                 dtype=np.float64,
                                                 shape=shape + (vec_size,),
                                                 dim_tags=",".join("c" for i in range(len(shape) + 1)),
                                                 base_storage=quantity + '_base_storage',
                                                 scope=lp.temp_var_scope.PRIVATE,
                                                 )

            def get_quantity_subscripts(e, zero=False):
                if isinstance(e, prim.Subscript):
                    index = e.index
                    if isinstance(index, tuple):
                        return index
                    else:
                        return (index,)
                else:
                    if zero:
                        return (0,)
                    else:
                        return ()

            for expr in quantities[quantity]:
                replacemap_vec[expr] = prim.Subscript(prim.Variable(get_vector_view_name(arrname)), get_quantity_subscripts(expr, zero=True) + (prim.Variable(new_iname),))

            for insn in write_insns:
                if isinstance(insn, lp.Assignment):
                    assignee = prim.Subscript(prim.Variable(arrname), get_quantity_subscripts(insn.assignee) + (prim.Variable('rotate_index'),))
                    new_insns.append(insn.copy(assignee=assignee,
                                               depends_on_is_final=True,
                                               )
                                     )
                elif isinstance(insn, lp.CInstruction):
                    # Rip apart the code and change the assignee
                    assignee, expression = insn.code.split("=")

                    # TODO This is a bit whacky: It only works for scalar assignees
                    # OTOH this code is on its way out anyway, because of CInstruction
                    assignee = prim.Subscript(prim.Variable(arrname), (prim.Variable('rotate_index'),))

                    code = "{} ={}".format(str(assignee), expression)
                    new_insns.append(insn.copy(code=code,
                                               depends_on_is_final=True,
                                               ))
                else:
                    raise NotImplementedError
        elif quantity in knl.temporary_variables:
            # Add a vector view to this quantity
            knl = add_vector_view(knl, quantity)
            replacemap_vec[expr] = prim.Subscript(prim.Variable(get_vector_view_name(quantity)),
                                                  (prim.Sum((prim.FloorDiv(prim.Variable("total_index"), vec_size), -1)), prim.Variable(new_iname)),
                                                  )

    other_insns = [i for i in knl.instructions if i.id not in [j.id for j in insns + new_insns]]

    #
    # Add two counter variables to the kernel
    #

    temporaries = knl.temporary_variables

    # Insert a flat consecutive counter 'total_index'
    temporaries['total_index'] = lp.TemporaryVariable('total_index',  # name
                                                      dtype=np.int32,
                                                      scope=lp.temp_var_scope.PRIVATE,
                                                      )
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_total_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   prim.Sum((prim.Variable("total_index"), 1)),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset(all_writers).union(frozenset({"assign_total_index"})),
                                   depends_on_is_final=True,
                                   id="update_total_index",
                                   ))

    # Insert a rotating index, that counts 0 , .. , vecsize - 1
    temporaries['rotate_index'] = lp.TemporaryVariable('rotate_index',  # name
                                                       dtype=np.int32,
                                                       scope=lp.temp_var_scope.PRIVATE,
                                                       )
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   id="assign_rotate_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   prim.Remainder(prim.Sum((prim.Variable("rotate_index"), 1)), vec_size),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset(all_writers).union(frozenset({"assign_rotate_index"})),
                                   depends_on_is_final=True,
                                   id="update_rotate_index",
                                   ))

    knl = knl.copy(temporary_variables=temporaries)

    #
    # Add a continue statement depending on the rotate index
    #

    # Determine the condition for the continue statement
    upper_bound = prim.Product(tuple(pw_aff_to_expr(knl.get_iname_bounds(i).size) for i in inames))
    total_check = prim.Comparison(prim.Variable("total_index"), "<", upper_bound)
    rotate_check = prim.Comparison(prim.Variable("rotate_index"), "!=", 0)
    check = prim.LogicalAnd((rotate_check, total_check))

    # Insert the 'continue' statement
    new_insns.append(lp.CInstruction((),  # iname exprs that the code needs access to
                                     "continue;",  # the code
                                     predicates=frozenset({check}),
                                     depends_on=frozenset({"update_rotate_index", "update_total_index"}).union(frozenset(all_writers)),
                                     depends_on_is_final=True,
                                     within_inames=common_inames.union(inames),
                                     within_inames_is_final=True,
                                     id="continue_stmt",
                                     ))

    #
    # Reconstruct the compute instructions
    #

    for insn in insns:
        # Get a vector view of the lhs expression
        lhsname = get_pymbolic_basename(insn.assignee)
        knl = add_vector_view(knl, lhsname)
        lhsname = get_vector_view_name(lhsname)

        new_insns.append(lp.Assignment(prim.Subscript(prim.Variable(lhsname),
                                                      (prim.Sum((prim.FloorDiv(prim.Variable("total_index"), vec_size), -1)), prim.Variable(new_iname)),
                                                      ),
                                       substitute(insn.expression, replacemap_vec),
                                       depends_on=frozenset({"continue_stmt"}),
                                       depends_on_is_final=True,
                                       within_inames=common_inames.union(frozenset(inames + (new_iname,))),
                                       within_inames_is_final=True,
                                       id=insn.id,
                                       )
                         )

    return knl.copy(instructions=new_insns + other_insns)
